/*
 * Copyright (c) 2018, 2019, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#ifndef SHARE_MEMORY_METASPACE_ABSTRACTPOOL_HPP
#define SHARE_MEMORY_METASPACE_ABSTRACTPOOL_HPP

#include "memory/allocation.hpp"
#include "utilities/debug.hpp"
#include "utilities/globalDefinitions.hpp"

namespace metaspace {

// A simple helper class;
//
// holds a linear array of elements of type E;
//
// array lives in C heap and expands automatically;
//
// free elements can be returned and are kept in a freelist;
//
// elements can be retrieved by their index (index type I).
//
// E must offer a "E* next()" and "set_next(E*)" methods.
// I must be an integral type.
//
template <class E, class I>
class AbstractPool {
public:

  typedef I size_type_t;

private:

  E* _arr;
  size_type_t _len;
  size_type_t _used;
  E* _freelist;
  const char* const _name;

  // Enlarge internal array if needed. Zero out new portion of array.
  void enlarge_to(size_type_t new_len) {
    const size_t new_size_bytes = new_len * sizeof(E);
    E* p = (E*) os::realloc(_arr, new_size_bytes, mtInternal);
    if (p == NULL) {
      vm_exit_out_of_memory(new_size_bytes, OOM_MALLOC_ERROR, "Pool %s: not enough space", _name);
    }
    _arr = p; _len = new_len;
  }

#ifdef ASSERT
  void check_elem(E* p) { assert(is_valid_elem(p), "invalid pointer"); }
  void check_idx(I i) { assert(i <= _used, "invalid index"); }
#endif

public:

  AbstractPool(const char* name)
    : _arr(NULL), _len(0), _used(0), _name(name), _freelist(NULL) {}
  ~AbstractPool() { os::free(_arr); }

  I index_for_elem(E* p) const {
    DEBUG_ONLY(check_elem(p));
    return p - _arr;
  }

  E* elem_at_index(I i) const {
    DEBUG_ONLY(check_idx(i));
    return _arr + i;
  }

  // Allocate a new element. Enlarge internal array if needed.
  // Will return NULL if cap is reached.
  E* allocate_element() {

    if (_freelist != NULL) {
      E* p = _freelist;
      _freelist = _freelist->next();
      DEBUG_ONLY(p->set_next(NULL);)
      return p;
    }

    if (_len == _used) {
      size_type_t new_len = _len == 0 ? 0x100 : _len * 2;
      enlarge_to(new_len);
    }

    // Avoid handing out anything at index 0.
    // This allows callers to use "I index == 0" as "invalid ref".
    if (_used == 0) {
      _used ++;
    }

    E* p = _arr[_used ++];
    return p;

  }

  void return_element(E* p) {
    DEBUG_ONLY(check_elem(p));
    p->set_next(_freelist);
    _freelist = p;
  }

  // Returns true if p was allocated from this pool.
  bool is_valid_elem(const E* p) { return p >= _arr && p < _arr + _used; }

};

} // namespace metaspace

#endif // SHARE_MEMORY_METASPACE_ABSTRACTPOOL_HPP
