/*
 * Copyright (c) 2018, 2019, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#ifndef SHARE_MEMORY_METASPACE_CHUNKLEVEL_HPP
#define SHARE_MEMORY_METASPACE_CHUNKLEVEL_HPP

#include "utilities/globalDefinitions.hpp"

// Constants for the chunk levels and some utility functions.

namespace metaspace {

// Metachunk level
typedef unsigned chklvl_t;

// Chunks are managed by a quaternary buddy allocator (allowing four empty
// chunks to be combined to a larger chunk).

// Chunk sizes range from 1K to 4MB (64bit).
//
// Reasoning: .... TODO explain

// We use a quaternary buddy allocator instead of a simple binary one because
// binary trees would get too deep and require too much memory for the nodes;
// A 1:4 merge strategy is enough for our causes, and keeps the costs for the
// tree nodes cheap.

// Each chunk has a level; the level corresponds to its position in the tree
// and describes its size.
//
// The largest chunks are called root chunks, of 4MB in size, and have level 0.
// From there on it goes:
//
// size    level
// 4MB     0
// 1MB     1
// 256K    2
// 64K     3
// 16K     4
// 4K      5
// 1K      6

namespace chklvl {

static const size_t   MAX_CHUNK_BYTE_SIZE    = 4 * M;
static const int      NUM_CHUNK_LEVELS       = 7;
static const size_t   MIN_CHUNK_BYTE_SIZE    = (MAX_CHUNK_BYTE_SIZE >> (size_t)NUM_CHUNK_LEVELS);

static const size_t   MIN_CHUNK_WORD_SIZE    = MIN_CHUNK_BYTE_SIZE / sizeof(MetaWord);
static const size_t   MAX_CHUNK_WORD_SIZE    = MAX_CHUNK_BYTE_SIZE / sizeof(MetaWord);

static const chklvl_t ROOT_CHUNK_LEVEL       = 0;

static const chklvl_t HIGHEST_CHUNK_LEVEL    = NUM_CHUNK_LEVELS - 1;
static const chklvl_t LOWEST_CHUNK_LEVEL     = 0;

inline bool is_valid_level(chklvl_t level) {
  return level <= HIGHEST_CHUNK_LEVEL;
}

inline void check_valid_level(chklvl_t lvl) {
  assert(is_valid_level(lvl), "invalid level (%d)", (int)lvl);
}

// Given a level return the chunk size, in words.
inline size_t word_size_for_level(chklvl_t level) {
  assert(is_valid_level(level), "invalid chunk level (%d)", level);
  return MAX_CHUNK_BYTE_SIZE >> level;
}

// Given an arbitrary word size smaller than the highest chunk size,
// return the highest chunk level able to hold this size.
chklvl_t level_fitting_word_size(size_t word_size);

// Shorthands to refer to exact sizes
const chklvl_t CHUNK_LEVEL_4M =     ROOT_CHUNK_LEVEL;
const chklvl_t CHUNK_LEVEL_1M =    (ROOT_CHUNK_LEVEL + 1);
const chklvl_t CHUNK_LEVEL_256K =  (ROOT_CHUNK_LEVEL + 2);
const chklvl_t CHUNK_LEVEL_64K =   (ROOT_CHUNK_LEVEL + 3);
const chklvl_t CHUNK_LEVEL_16K =   (ROOT_CHUNK_LEVEL + 4);
const chklvl_t CHUNK_LEVEL_4K =    (ROOT_CHUNK_LEVEL + 5);
const chklvl_t CHUNK_LEVEL_1K =    (ROOT_CHUNK_LEVEL + 6);;

STATIC_ASSERT(CHUNK_LEVEL_1K == HIGHEST_CHUNK_LEVEL);

} // namespace chklvl

} // namespace metaspace

#endif // SHARE_MEMORY_METASPACE_BLOCKFREELIST_HPP
