/*
 * Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */
#include "precompiled.hpp"

//#include "logging/log.hpp"
#include "memory/metaspace/chunkTree.hpp"
#include "memory/metaspace/metachunk.hpp"
//#include "utilities/ostream.hpp"
#include "utilities/debug.hpp"
#include "utilities/globalDefinitions.hpp"

namespace metaspace {

ChunkTreeNodePool ChunkTree::_nodePool("metaspace chunk tree node pool");


#ifdef ASSERT
void ChunkTree::verify_node(const qtnode_t* n) {

  // Child pointers are never NULL; either one of them may point to another node,
  // or to a Metachunk*.
  assert(n->child[0] != NULL &&
         n->child[1] != NULL &&
         n->child[2] != NULL &&
         n->child[3] != NULL, "child ptrs null");

  if (n->parent != NULL) {
    _nodePool.check_elem(n->parent);
  }

  for (int i = 0; i < 4; i ++) {
    // Each child pointer shall point to either another node or to a chunk header
    Metachunk* c = resolve_chunk_ptr(n->child[i]);
    qtnode_t* n2 = resolve_node_ptr(n->child[i]);
    if (c != NULL) {
      // It is a chunk.
      assert(c->tree_node() == n, "Chunk must have backlink to parent");
      assert(n2 == NULL, "only one");
    } else {
      // It is a node.
      assert(n2->parent == n, "Chunk must have backlink to parent");
      assert(c == NULL, "only one");
    }
  }

}
#endif


// Initialize: allocate a root node and a root chunk header; return the
// root chunk header.
Metachunk* ChunkTree::alloc_root() {

  assert(_root == NULL, "already have a root");

  Metachunk* c = MetachunkPool::alloc_header();

  // Newly created chunks should have level==0 aka root
  assert(c->level() == chklvl::ROOT_CHUNK_LEVEL, "not root chunk?");

  _root = c;

  return c;

}


// Given a chunk c, split it into four smaller chunks.
// Returns pointer to the result chunk; appends splintered off chunks to the given linked list.
// Returns NULL if chunk cannot be split.
Metachunk* ChunkTree::split_once(Metachunk* c, Metachunk** p_splinters) {

  Metachunk* remainder_list = NULL;

  if (c->level() == chklvl::HIGHEST_CHUNK_LEVEL) {
    return NULL;
  }

  MetaWord* new_base = c->base();
  const int new_level = c->level() + 1;
  const size_t new_word_size = chklvl::word_size_for_level(new_level);
  qtnode_t* const old_parent = c->tree_node();

  // 1. Create a single new parent node...
  qtnode_t* new_parent_node = allocate_node();

  // 2. the originating chunk header is reused as the first chunk header
  //    of all splinters.
  c->set_level(new_level);
  c->set_tree_node(new_parent_node);
  new_parent_node->child[0] = c;

  // 3. Create three new sibling chunk headers; initialize them;
  //    and append them to the splinters list.
  for (int i = 0; i < 3; i ++) {
    Metachunk* sib = MetachunkPool::alloc_header();
    sib->set_level(new_level);
    sib->set_base(new_base);

    new_base += new_word_size;

    sib->set_tree_node(new_parent_node);
    new_parent_node->child[i] = sib;

    // Return splinters in this linked list
    sib->set_next(*p_splinters);
    *p_splinters = sib;
  }

  // 4. ... Now, replace the old header storage location with the
  //        new parent node.
  if (old_parent == NULL) {
    assert(_root == c, "Must have been root");
    _root = new_parent_node;
  } else {
    if (old_parent->child[0] == c) {
      old_parent->child[0] = new_parent_node;
    } else if (old_parent->child[1] == c) {
      old_parent->child[1] = new_parent_node;
    } else if (old_parent->child[2] == c) {
      old_parent->child[2] = new_parent_node;
    } else if (old_parent->child[3] == c) {
      old_parent->child[3] = new_parent_node;
    } else {
      ShouldNotReachHere();
    }
  }

  return c;
}

// Given a chunk c, split it repeatedly until you get a chunk of the given target_level.
// Return the remainder chunks in a linked list.
Metachunk* ChunkTree::split(chklvl_t target_level, Metachunk* c, Metachunk** p_splinters) {

  DEBUG_ONLY(chklvl::check_valid_level(target_level));
  assert(target_level > c->level(), "Wrong target level");

  Metachunk* c_last = split_once(c, p_splinters);
  while (c_last != NULL && c_last->level() < target_level) {
    c_last = split_once(c, p_splinters);
  }

  assert(c != NULL, "Sanity");

  return c;

}


// Given a chunk, attempt to merge it with its siblings
// if they are free.
// Returns pointer to the result chunk if successful, NULL otherwise.
//
// !!! Please note that if this method returns a non-NULL value, the
// original chunk will be invalid and should not be accessed anymore! !!!
Metachunk* ChunkTree::merge_once(Metachunk* c) {

  DEBUG_ONLY(c->verify(false);)
  assert(c->is_free(), "Only call for free chunks.");

  if (c->level() == chklvl::ROOT_CHUNK_LEVEL) {
    return NULL;
  }

  qtnode_t* old_parent = c->tree_node();
  assert(old_parent != NULL, "Sanity");

  // Are all neighbors able to merge? They are if none is splintered (they are
  // all leaf nodes) and free.
  Metachunk* s0, *s1, *s2, *s3;
  if ((s0 = resolve_chunk_ptr(old_parent->child[0])) == NULL || s0->is_in_use() ||
      (s1 = resolve_chunk_ptr(old_parent->child[1])) == NULL || s1->is_in_use() ||
      (s2 = resolve_chunk_ptr(old_parent->child[2])) == NULL || s2->is_in_use() ||
      (s3 = resolve_chunk_ptr(old_parent->child[3])) == NULL || s3->is_in_use()) {
    return NULL;
  }

  // The new merged chunk is the first one. Nothing but the level changes.
  Metachunk* c_merged = s0;
  const int new_level = c->level() - 1;
  c_merged->set_level(new_level);

  // The other chunk headers are not needed anymore and can be returned to the
  // pool.
  MetachunkPool::release_header(s1);
  MetachunkPool::release_header(s2);
  MetachunkPool::release_header(s3);

  // The former parent node is not needed anymore. Add the new chunk to its parent.
  qtnode_t* old_grandparent = old_parent->parent;
  if (old_grandparent == NULL) {
    // old parent was root.
    _root = c_merged;
  } else {
    // old parent was not NULL. Replace the pointer to old parent in the grand parent
    // child array with the newly created chunk header.
    if (old_grandparent->child[0] == old_parent) {
      old_grandparent->child[0] = c_merged;
    } else if (old_grandparent->child[1] == old_parent) {
      old_grandparent->child[1] = c_merged;
    } else if (old_grandparent->child[2] == old_parent) {
      old_grandparent->child[2] = c_merged;
    } else if (old_grandparent->child[3] == old_parent) {
      old_grandparent->child[3] = c_merged;
    } else {
      ShouldNotReachHere();
    }
  }

  // Old parent node is not needed anymore
  release_node(old_parent);

  // Now repeat recursively:
  return c_merged;

}


// Given a chunk, attempt to merge it with its neighboring chunks
// if they are free.
// Does so recursively until either the chunk is a root chunk or
// one of the neighbors is non-free.
// If successful (at least one merge happened), returns address of
// the merged chunk; NULL otherwise.
//
// !!! Please note that if this method returns a non-NULL value, the
// original chunk will be invalid and should not be accessed anymore! !!!
Metachunk* ChunkTree::merge(Metachunk* c) {

  DEBUG_ONLY(c->verify(false);)
  assert(c->is_free(), "Only merge free chunks.");
  assert(c->level() > chklvl::ROOT_CHUNK_LEVEL, "Only merge root chunks.");

  Metachunk* c_last = merge_once(c);
  while (c_last != NULL && c_last->level() > chklvl::ROOT_CHUNK_LEVEL) {
    c_last = merge_once(c);
  }

  return c_last;
}


//// ChunkTreeArray /////////////777

// Create an array of ChunkTree objects, all initialized to NULL, covering
// a given memory range. Memory range must be aligned to size of root chunks.
ChunkTreeArray::ChunkTreeArray(const MetaWord* base, size_t word_size)
  : _base(base), _word_size(word_size),
    _arr(NULL), _num(0)
{
  assert(is_aligned(_word_size, chklvl::MAX_CHUNK_WORD_SIZE), "not aligned");
  _num = _word_size / chklvl::MAX_CHUNK_WORD_SIZE;
  _arr = NEW_C_HEAP_ARRAY(ChunkTree*, _num, mtInternal);
  memset(_arr, 0, sizeof(ChunkTree*) * _num);
}

ChunkTreeArray::~ChunkTreeArray() {
  FREE_C_HEAP_ARRAY(ChunkTree*, _arr);
}

} // namespace metaspace


