/*
 * Copyright (c) 2018, 2019, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#ifndef SHARE_MEMORY_METASPACE_CHUNKTREE_HPP
#define SHARE_MEMORY_METASPACE_CHUNKTREE_HPP

#include "memory/metaspace/abstractPool.hpp"
#include "memory/metaspace/chunkLevel.hpp"
#include "memory/metaspace/metachunk.hpp" // For MetachunkPool

namespace metaspace {

// Chunks live in a quaternary tree.
//
// Each node has four child nodes.
// Leaf nodes are the chunk headers.

struct qtnode_t;

struct qtnode_t {

  // parent node; NULL if this is a root node.
  qtnode_t* parent;

  // Child nodes can be either non-leaf itself, in which case they are qtnode_t*;
  // or they are leaf nodes, in which case they are Metachunk*.
  void* child[4];

  // For the purpose of keeping this structure in a free list of the
  // ChunkTreeNodePool, reuse the parent pointer as next (this only
  // happens to dead nodes, so the parent pointer is not needed).
  // See AbstractPool<> class.
  qtnode_t* next() { return parent; }
  void set_next(qtnode_t* n) { parent = n; }

#ifdef ASSERT
  bool check_is_child(const void* p) const {
    return child[0] == p || child[1] == p || child[2] == p || child[3] == p;
  }
#endif

};

typedef AbstractPool<qtnode_t, uint32_t> ChunkTreeNodePool;

class ChunkTree {

  // Root is either a direct pointer to a Metachunk* (in that case, a root chunk of max. size)
  // or a pointer to a node.
  void* _root;

  /// Node allocation ///

  static ChunkTreeNodePool _nodePool;

  // Returns a new node. Node will be zero'd out completely.
  static qtnode_t* allocate_node() {
    qtnode_t* n = _nodePool.allocate_element();
    memset(n, 0, sizeof(n));
    return n;
  }

  static qtnode_t* release_node(qtnode_t* n) {
    return _nodePool.return_element(n);
  }

  // Given a pointer to a child which can either be a Metachunk or a node, return a pointer to a
  // node if it was a node; NULL if it was a Metachunk.
  static qtnode_t* resolve_node_ptr(void* p) {
    qtnode_t* n = (qtnode_t*)p;
    if (_nodePool.is_valid_elem(n)) {
      return n;
    }
    return NULL;
  }

  // Given a pointer to a child which can either be a Metachunk or a node, return a pointer to a
  // node if it was a node; NULL if it was a Metachunk.
  static Metachunk* resolve_chunk_ptr(void* p) {
    Metachunk* c = (Metachunk*) p;
    if (MetachunkPool::is_metachunk_pointer(c)) {
      return c;
    }
    return NULL;
  }

#ifdef ASSERT
  // Verify a life node (one which lives in the tree).
  static void verify_node(const qtnode_t* n);
  static void verify_parenthood(const qtnode_t* child, const qtnode_t* parent);
  // Given four leaf nodes, verify their chunks (must be adjacent and in order)
  static void verify_sibling_chunks(const Metachunk* sibs[4]);
#endif


  // Given a chunk c, split it into four smaller chunks.
  // Returns pointer to the result chunk; splintered off chunks are appended to the given linked list.
  // Returns NULL if chunk cannot be split.
  Metachunk* split_once(Metachunk* c, Metachunk** p_splinters);

  // Given a chunk, attempt to merge it with its siblings
  // if they are free.
  // Returns pointer to the result chunk if successful, NULL otherwise.
  //
  // !!! Please note that if this method returns a non-NULL value, the
  // original chunk will be invalid and should not be accessed anymore! !!!
  Metachunk* merge_once(Metachunk* c);


public:

  ChunkTree() : _root(NULL) {}

  // Initialize: allocate a root node and a root chunk header; return the
  // root chunk header.
  Metachunk* alloc_root();

  // Given a chunk c, split it recursively until you get a chunk of the given target_level.
  // Returns pointer to the result chunk; returns split off chunks in p_splinters as linked list.
  // Returns NULL if chunk cannot be split at least once.
  Metachunk* split(chklvl_t target_level, Metachunk* c, Metachunk** p_splinters);

  // Given a chunk, attempt to merge it recursively with its neighboring chunks.
  // If successful (merged at least once), returns address of
  // the merged chunk; NULL otherwise.
  //
  // !!! Please note that if this method returns a non-NULL value, the
  // original chunk will be invalid and should not be accessed anymore! !!!
  Metachunk* merge(Metachunk* c);

  //// tree traversal ////

  class ConstNodeClosure {
   public:
    // Return false to cancel traversal.
    virtual bool do_node(const qtnode_t* node) = 0;
  };

  class ConstChunkClosure {
   public:
    // Return false to cancel traversal.
    virtual bool do_chunk(const Metachunk* c) = 0;
  };


  // Iterate over all leafs in this tree. Returns true for complete traversal,
  // false if traversal was cancelled.
  bool iterate_nodes(ConstNodeClosure* c) const;

  // Iterate over all nodes in this tree. Returns true for complete traversal,
  // false if traversal was cancelled.
  bool iterate_chunks(ConstChunkClosure* c) const;


  //// Debug stuff ////

  DEBUG_ONLY(void verify(bool slow);)

};


///////////////////////
// An C-heap allocated array of chunk trees. Used to describe fragmentation over a range of multiple root chunks.
class ChunkTreeArray {

  const MetaWord* const _base;
  const size_t _word_size;

  ChunkTree* _arr;
  int _num;

#ifdef ASSERT
  void check_pointer(const MetaWord* p) const {
    assert(p >= _base && p < _base + _word_size, "Invalid pointer");
  }
#endif

  int index_by_address(const MetaWord* p) const {
    DEBUG_ONLY(check_pointer(p);)
    return (p - _base) / chklvl::MAX_CHUNK_WORD_SIZE;
  }

public:

  // Create an array of ChunkTree objects, all initialized to NULL, covering
  // a given memory range. Memory range must be aligned to size of root chunks.
  ChunkTreeArray(const MetaWord* base, size_t word_size);

  ~ChunkTreeArray();

  // Given a memory address into the range the trees cover, return the corresponding
  // tree. If none existed at this position, create it.
  ChunkTree* get_tree_by_address(const MetaWord* p) const {
    assert(p >= _base && p < _base + _word_size, "Invalid pointer");
    const int idx = index_by_address(p);
    assert(idx < _num, "Invalid index");
    if (_arr[idx] == NULL) {
      _arr[idx] = new ChunkTree();
    }
    return _arr[idx];
  }

};


} // namespace metaspace

#endif // SHARE_MEMORY_METASPACE_CHUNKTREE_HPP
