/*
 * Copyright (c) 2012, 2016, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */


#include "precompiled.hpp"

#include "memory/metaspace/chunkLevel.hpp"
#include "memory/metaspace/chunkTree.hpp"
#include "memory/metaspace/constants.hpp"
#include "memory/metaspace/metachunk.hpp"
#include "memory/metaspace/virtualSpaceNode.hpp"

#include "utilities/align.hpp"
#include "utilities/copy.hpp"
#include "utilities/debug.hpp"

namespace metaspace {

// Must align pointers and sizes to 8 so that 64 bit types get correctly aligned.
static const size_t allocation_alignment = 8;

// Make sure that the Klass alignment also agree.
STATIC_ASSERT(allocation_alignment == (size_t)KlassAlignmentInBytes);

void Metachunk::remove_from_list() {
  if (_prev != NULL) {
    _prev->set_next(_next);
  }
  if (_next != NULL) {
    _next->set_prev(_prev);
  }
  _prev = _next = NULL;
}

// Allocate word_size words from this chunk.
//
// May cause memory to be committed. That may fail if we hit a commit limit. In that case,
//  NULL is returned and p_did_hit_commit_limit will be set to true.
// If the remainder portion of the chunk was too small to hold the allocation,
//  NULL is returned and p_did_hit_commit_limit will be set to false.
MetaWord* Metachunk::allocate(size_t word_size, bool* p_did_hit_commit_limit) {

  size_t request_word_size = align_up(word_size, allocation_alignment);

  // Space enough left?
  if (free_words() < request_word_size) {
    *p_did_hit_commit_limit = false;
    return NULL;
  }

  // Will bring this allocation us over the border of the next commit granule?
  MetaWord* const top = top();
  MetaWord* const next_top = top() + request_word_size;

  MetaWord* const commit_granule_border = align_up(top, constants::commit_granule_words);

  if (next_top > commit_granule_border) {
    // We may need to commit.

    // Calc the extend of the area we may need to commit.
    MetaWord* const commit_granule_border_2 = align_up(next_top, constants::commit_granule_words);

    // and attempt to commit.
    const size_t word_size_to_commit = commit_granule_border_2 - commit_granule_border;
    if (_vsnode->ensure_range_is_committed(commit_granule_border, word_size_to_commit) == false) {
      // ... which failed because we hit a limit.
      *p_did_hit_commit_limit = true;
      return NULL;
    }

  }

  // All is well.
  _used_words += request_word_size;

  return top;

}

#ifdef ASSERT

void Metachunk::verify(bool slow) const {

  // Note: only call this on a life Metachunk.

  chklvl::check_valid_level(level());

  // Test base pointer
  assert(vsnode() != NULL, "No space");
  vsnode()->check_pointer(base());
  assert(base() != NULL, "Base pointer NULL");

  // Starting address shall be aligned to chunk size.
  const size_t required_alignment = word_size() * sizeof(MetaWord);
  assert(is_aligned(base(), required_alignment),
         "Chunk " PTR_FORMAT ": (size " SIZE_FORMAT ") not aligned correctly to " SIZE_FORMAT ".",
         p2i(this), word_size() * sizeof(MetaWord), required_alignment);

  // Used words
  assert(used_words() < word_size(), "oob");

  // We shall have a reference to a tree node and that node shall refer back to us
  assert(tree_node() != NULL &&
         tree_node()->check_is_child(this), "Invalid tree node relationship");


}
#endif // ASSERT

} // namespace metaspace

