/*
 * Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#include <memory/metaspace/commitCounter.hpp>
#include "precompiled.hpp"

#include "logging/log.hpp"

#include "memory/metaspace/constants.hpp"
#include "memory/metaspace/metachunk.hpp"
#include "memory/metaspace/virtualSpaceNode.hpp"

#include "runtime/mutexLocker.hpp"
#include "runtime/os.hpp"

#include "utilities/align.hpp"
#include "utilities/debug.hpp"
#include "utilities/globalDefinitions.hpp"

namespace metaspace {

#ifdef ASSERT
template <class T>
void check_is_aligned_to_commit_granule(T x) {
  assert(is_aligned(x, constants::commit_granule_bytes), "Unaligned pointer");
}
#endif

// Given an address range (which has to be aligned to commit granule size):
//  - commit it
//  - mark it as committed in the commit mask
// May fail and return NULL if we hit a global commit limit (GC threshold, MaxMetaspaceSize).
// Will assert if we run out of memory.
bool VirtualSpaceNode::commit_range(MetaWord* p, size_t word_size) {

  DEBUG_ONLY(check_is_aligned_to_commit_granule(p);)
  DEBUG_ONLY(check_is_aligned_to_commit_granule(word_size);)
  assert_lock_strong(MetaspaceExpand_lock);

  // First calculate how large the committed regions in this range are
  const size_t committed_words_in_range = _commit_mask.get_committed_size_in_range(p, word_size);
  DEBUG_ONLY(check_is_aligned_to_commit_granule(committed_words_in_range);)

  // By how much words we would increase commit charge
  //  were we to commit the given address range completely.
  const size_t commit_increase_words = word_size - committed_words_in_range;

  if (commit_increase_words == 0) {
    return true; // Already fully committed, nothing to do.
  }

  // Before committing any more memory, check limits.
  if (CommitCounter::possible_expansion_words() < commit_increase_words) {
    return false;
  }

  // Commit...
  if (os::commit_memory((char*)p, word_size * BytesPerWord, false) == false) {
    vm_exit_out_of_memory(word_size * BytesPerWord, OOM_MMAP_ERROR, "Failed to commit metaspace.");
  }

  log_debug(gc, metaspace)("Increased metaspace by " SIZE_FORMAT " bytes.",
                           commit_increase_words * BytesPerWord);

  // ... increase global commit counter..
  CommitCounter::increase_committed(commit_increase_words);

  // ... and update the commit mask.
  _commit_mask.mark_range_as_committed(p, word_size);

  return true;

}

// Given an address range (which has to be aligned to commit granule size), ensure
// it is committed.
//  - commit it
//  - mark it as committed in the commit mask
// May fail and return NULL if we hit a global commit limit (GC threshold, MaxMetaspaceSize).
// Will assert if we run out of memory.
bool VirtualSpaceNode::ensure_range_is_committed(MetaWord* p, size_t word_size) {

  // Todo: simple for now. Make it more intelligent late
  return commit_range(p, word_size);

}

// Given an address range (which has to be aligned to commit granule size):
//  - uncommit it
//  - mark it as uncommitted in the commit mask
bool VirtualSpaceNode::uncommit_range(MetaWord* p, size_t word_size) {

  DEBUG_ONLY(check_is_aligned_to_commit_granule(p);)
  DEBUG_ONLY(check_is_aligned_to_commit_granule(word_size);)
  assert_lock_strong(MetaspaceExpand_lock);

  // First calculate how large the committed regions in this range are
  const size_t committed_words_in_range = _commit_mask.get_committed_size_in_range(p, word_size);
  DEBUG_ONLY(check_is_aligned_to_commit_granule(committed_words_in_range);)

  if (committed_words_in_range == 0) {
    return true; // Already fully uncommitted, nothing to do.
  }

  // Uncommit...
  if (os::uncommit_memory((char*)p, word_size * BytesPerWord) == false) {
    // Note: this can actually happen, since uncommit may increase the number of mappings.
    fatal("Failed to uncommit metaspace.");
  }

  log_debug(gc, metaspace)("Decreased metaspace by " SIZE_FORMAT " bytes.",
                            committed_words_in_range * BytesPerWord);

  // ... increase global commit counter..
  CommitCounter::decrease_committed(committed_words_in_range);

  // ... and update the commit mask.
  _commit_mask.mark_range_as_uncommitted(p, word_size);

  return true;

}

//// creation ////

// Create a new empty node spanning the given reserved space.
VirtualSpaceNode::VirtualSpaceNode(ReservedSpace rs)
  : _next(NULL)
  , _rs(rs)
  , _base(rs.base())
  , _word_size(rs.size() / BytesPerWord)
  , _used_words(0)
  , _commit_mask(rs.base(), rs.size() / BytesPerWord)
  , _chunk_tree_array(rs.base(), rs.size() / BytesPerWord)
{}

// Create a node of a given size
static VirtualSpaceNode* VirtualSpaceNode::create_node(size_t word_size) {

  DEBUG_ONLY(check_is_aligned_to_commit_granule(word_size);)

  ReservedSpace rs(word_size * BytesPerWord,
                   constants::commit_granule_bytes,
                   false, // TODO deal with large pages
                   false);

  if (!rs.is_reserved()) {
    vm_exit_out_of_memory(word_size * BytesPerWord, OOM_MMAP_ERROR, "Failed to reserve memory for metaspace");
  }

  return create_node(rs);

}

// Create a node over an existing space
static VirtualSpaceNode* VirtualSpaceNode::create_node(ReservedSpace rs) {
  return new VirtualSpaceNode(rs);
}



//// Chunk allocation, splitting, merging /////

// Allocate a root chunk from this node. Will fail and return NULL
// if the node is full.
// Note that the chunk memory may or may not be committed.
Metachunk* VirtualSpaceNode::allocate_root_chunk() {

  const size_t remaining_words = _word_size - _used_words;

  if (remaining_words >= chklvl::MAX_CHUNK_WORD_SIZE) {

    MetaWord* loc = _base + _used_words;
    _used_words += chklvl::MAX_CHUNK_WORD_SIZE;

    // Create a new chunk tree for that new root node.
    ChunkTree* tree = _chunk_tree_array.get_tree_by_address(loc);

    // Create a root chunk and initialize it;
    Metachunk* c = tree->alloc_root();
    c->set_base(loc);

    DEBUG_ONLY(c->verify(true);)
    return c;

  }

  // There should be no "leftover" space since the mapping for a
  // VirtualSpaceNode should always be aligned to root chunk size.
  assert(remaining_words == 0, "Sanity");

  return NULL; // Node is full.

}

// Given a chunk c, split it recursively until you get a chunk of the given target_level.
// Returns pointer to the result chunk; returns split off chunks in p_splinters as linked list.
// Returns NULL if chunk cannot be split at least once.
Metachunk* VirtualSpaceNode::split(chklvl_t target_level, Metachunk* c, Metachunk** p_splinters) {

  // Get the tree associated with this chunk and let it handle the splitting
  ChunkTree* tree = _chunk_tree_array.get_tree_by_address(c->base());
  return tree->split(target_level, c, p_splinters);

}

// Given a chunk, attempt to merge it recursively with its neighboring chunks.
// If successful (merged at least once), returns address of
// the merged chunk; NULL otherwise.
//
// !!! Please note that if this method returns a non-NULL value, the
// original chunk will be invalid and should not be accessed anymore! !!!
Metachunk* VirtualSpaceNode::merge(Metachunk* c) {

  // Get the tree associated with this chunk and let it handle the merging
  ChunkTree* tree = _chunk_tree_array.get_tree_by_address(c->base());
  return tree->merge(c);

}

#ifdef ASSERT
// Verify counters and basic structure. Slow mode: verify all chunks in depth
void VirtualSpaceNode::verify(bool slow) {



}
#endif


} // namespace metaspace

