/*
 * Copyright (c) 2018, 2019, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#ifndef SHARE_MEMORY_METASPACE_VIRTUALSPACENODE_HPP
#define SHARE_MEMORY_METASPACE_VIRTUALSPACENODE_HPP

#include "memory/metaspace/constants.hpp"
#include "memory/metaspace/chunkTree.hpp"
#include "memory/metaspace/commitMask.hpp"
#include "memory/virtualspace.hpp"
#include "memory/memRegion.hpp"
#include "utilities/debug.hpp"
#include "utilities/bitMap.hpp"
#include "utilities/globalDefinitions.hpp"

class outputStream;

namespace metaspace {

// VirtualSpaceNode manage a single address range of the Metaspace.
//
// That address range may contain interleaved committed and uncommitted
// regions. It keeps track of which regions have committed and offers
// functions to commit and uncommit regions.
//
// It allocates and hands out memory ranges, starting at the bottom.
//
// Address range must be aligned to root chunk size.
//
class VirtualSpaceNode : public CHeapObj<mtClass> {

  // Link to next VirtualSpaceNode
  VirtualSpaceNode* _next;

  ReservedSpace _rs;

  // Start pointer of the area.
  MetaWord* const _base;

  // Size, in words, of the whole node
  const size_t _word_size;

  // Size, in words, of the used portion of the node
  size_t _used_words;

  // The bitmap describing the commit state of the region:
  // Each bit covers a region of 64K (see constants::commit_granule_size).
  CommitMask _commit_mask;

  // An array of chunk trees. Each one describes fragmentation inside the associated root chunk.
  ChunkTreeArray _chunk_tree_array;

  /// committing, uncommitting ///

  // Given a pointer into this node, calculate the start of the commit granule
  // the pointer points into.
  MetaWord* calc_start_of_granule(MetaWord* p) const {
    DEBUG_ONLY(check_pointer(p));
    return align_down(p, constants::commit_granule_bytes);
  }

  // Given an address range (which has to be aligned to commit granule size):
  //  - commit it
  //  - mark it as committed in the commit mask
  // May fail and return NULL if we hit a global commit limit (GC threshold, MaxMetaspaceSize).
  // Will assert if we run out of memory.
  bool commit_range(MetaWord* p, size_t word_size);

  //// creation ////

  // Create a new empty node spanning the given reserved space.
  VirtualSpaceNode(ReservedSpace rs);

public:

  // Create a node of a given size
  static VirtualSpaceNode* create_node(size_t word_size);

  // Create a node over an existing space
  static VirtualSpaceNode* create_node(ReservedSpace rs);

  ~VirtualSpaceNode();

  //// Chunk allocation, splitting, merging /////

  // Allocate a root chunk from this node. Will fail and return NULL
  // if the node is full.
  // Note that the chunk memory may or may not be committed. Commit its payload
  // area before using it.
  Metachunk* allocate_root_chunk();

  // Given a chunk c, split it recursively until you get a chunk of the given target_level.
  // Returns pointer to the result chunk; returns split off chunks in p_splinters as linked list.
  // Returns NULL if chunk cannot be split at least once.
  Metachunk* split(chklvl_t target_level, Metachunk* c, Metachunk** p_splinters);

  // Given a chunk, attempt to merge it recursively with its neighboring chunks.
  // If successful (merged at least once), returns address of
  // the merged chunk; NULL otherwise.
  //
  // !!! Please note that if this method returns a non-NULL value, the
  // original chunk will be invalid and should not be accessed anymore! !!!
  Metachunk* merge(Metachunk* c);


  //// Committing/uncommitting memory /////

  // Given an address range (which has to be aligned to commit granule size), ensure
  // it is committed.
  //  - commit it
  //  - mark it as committed in the commit mask
  // May fail and return NULL if we hit a global commit limit (GC threshold, MaxMetaspaceSize).
  // Will assert if we run out of memory.
  bool ensure_range_is_committed(MetaWord* p, size_t word_size);

  // Given an address range (which has to be aligned to commit granule size):
  //  - uncommit it
  //  - mark it as uncommitted in the commit mask
  bool uncommit_range(MetaWord* p, size_t word_size);


  /// Debug stuff ////

  // Verify counters and basic structure. Slow mode: verify all chunks in depth
  bool contains(const MetaWord* p) const {
    return p >= _base && p < _base + _used_words;
  }

#ifdef ASSERT
  void check_pointer(const MetaWord* p) const {
    assert(contains(p), "invalid pointer");
  }
  void verify(bool slow);
#endif

};


} // namespace metaspace

#endif // SHARE_MEMORY_METASPACE_VIRTUALSPACENODE_HPP
